#!/usr/bin/perl
#
#  Agent login/logoff 
#
#
# TODO: na primeira abertura abrir janela para configurar

use strict;
use warnings;
use threads;
use threads::shared;


use Thread::Semaphore;
use Thread::Queue;
use Config::Tiny;
use Data::Dumper;
use Storable qw(freeze thaw);
use Gtk2 -init;
use IO::Socket::INET;
use IO::Select;

my $quit : shared = 0;
# Nao conseguir usar cond_wait e cond_signal, usar semaphoro entao
my $main_window_mutex = Thread::Semaphore->new();
my $main_window_inited : shared = 0;

# filas de comunicacao entre threads
my $asterisk_in = Thread::Queue->new();
my $asterisk_out = Thread::Queue->new();

# thread para o manager com o asterisk 
my $thr_asterisk = threads->new(\&asterisk_ctl);
# thread para a janela principal
my $thr_window  = threads->new(\&main_window);

# esperar a janela ser desenhada (Cond_wait nao funcionou)
while(1) {
    $main_window_mutex->down();
    if ($main_window_inited == 1) {
        $main_window_mutex->up();
        last;
    }
    $main_window_mutex->up();
    # dormir por 0.1 segundos
    select(undef, undef, undef, 0.1);
}


Gtk2->main;

$thr_window->join();
$thr_asterisk->join();

sub main_window() {
    my %widgets;
    $widgets{"window_main"} = Gtk2::Window->new();
    
    $widgets{"window_main"}->signal_connect(destroy => \&fechar_tudo, \%widgets);
    $widgets{"window_main"}->set_title("O b e l i s k");

    $widgets{"hbox"} = Gtk2::HBox->new(0, 0);
    
    $widgets{"window_main"}->add($widgets{"hbox"});

    $widgets{"entry_bina"} = Gtk2::Entry->new();
    $widgets{"entry_bina"}->set_editable(0);
    $widgets{"entry_bina"}->set_width_chars(30);
    $widgets{"hbox"}->pack_start($widgets{"entry_bina"}, 1, 1, 0);

    $widgets{"vbox"} = Gtk2::VBox->new(0, 0);

    $widgets{"hbox"}->pack_end($widgets{"vbox"}, 0, 0, 0);
    
    $widgets{"button_pausar"} = Gtk2::Button->new_with_label("Login");
    $widgets{"button_pausar"}->signal_connect(clicked => \&pausar_despausar, \%widgets);
    $widgets{"vbox"}->pack_start($widgets{"button_pausar"}, 1, 0, 0);

    $widgets{"button_configurar"} = Gtk2::Button->new_with_label("Configurar");
    $widgets{"button_configurar"}->signal_connect(clicked => \&configurar, \%widgets);
    $widgets{"vbox"}->pack_start($widgets{"button_configurar"}, 1, 0, 0);

    $widgets{"window_main"}->show_all();

    $main_window_mutex->down();
    $main_window_inited = 1;
    $main_window_mutex->up();
    
    while (not $quit) {
        if(my $e = thaw($asterisk_out->dequeue_nb())) {
            if ($e->{TYPE} eq 'LOG') {
                #print Dumper($e);
                $widgets{"entry_bina"}->set_text($e->{DATA});
                #$asterisk_out->enqueue(freeze {TYPE => 'CLEAN'});
            } elsif ($e->{TYPE} eq 'BINA') {
                #print Dumper($e);
                $widgets{"entry_bina"}->set_text($e->{DATA});
            } elsif ($e->{TYPE} eq 'CLEAN') {
                if ($widgets{"button_pausar"}->get_label() eq 'Pausar') {
                    $widgets{"entry_bina"}->set_text("Agente Livre");
                } elsif ($widgets{"button_pausar"}->get_label() eq 'Despausar') {
                    $widgets{"entry_bina"}->set_text("Agente Pausado");
                } else {
                    $widgets{"entry_bina"}->set_text("");
                }
            } elsif ($e->{TYPE} eq 'INFO') {
                $widgets{"window_main"}->set_title("Obelisk - " . join("@", $e->{DATA}->{agent}, $e->{DATA}->{extension}));
            }   
        }
        select(undef, undef, undef, 0.1);
    
    };

}

sub asterisk_ctl {
    my $config_file = join("/", $ENV{HOME}, ".obeliskrc");
    my $erro;
    if (not -r $config_file) {
        # FIXME: nao ficar testando $erro para saber se pode continuar
        # nao sei fechar todas as threads de vez, die nao funciona, corrigir isto.
        $erro = "File not found or not readable $config_file\n";
    }
    my $server;
    my $port;
    my $username;
    my $secret;

    my $agent;
    my $queue;
    my $extension;
    my $channel;
    
    if (not $erro) {
        my $Config = Config::Tiny->read($config_file) if not $erro;
        
        if (
            not exists $Config->{MANAGER} or
            not exists $Config->{MANAGER}->{server} or
            not exists $Config->{MANAGER}->{port} or
            not exists $Config->{MANAGER}->{username} or
            not exists $Config->{MANAGER}->{secret} or
            not exists $Config->{AGENT} or
            not exists $Config->{AGENT}->{agent} or
            not exists $Config->{AGENT}->{queue} or
            not exists $Config->{AGENT}->{extension} or
            not exists $Config->{AGENT}->{channel}
        ) {
            $erro = "Error while loading the config file\n";
        } else {
            $server   = $Config->{MANAGER}->{server};
            $port     = $Config->{MANAGER}->{port};
            $username = $Config->{MANAGER}->{username};
            $secret   = $Config->{MANAGER}->{secret};
            
            $agent     = $Config->{AGENT}->{agent};
            $queue     = $Config->{AGENT}->{queue};
            $extension = $Config->{AGENT}->{extension};
            $channel   = $Config->{AGENT}->{channel};
        }

    }
    if ($erro) {
        # esperar a janela ser desenhada (Cond_wait nao funcionou)
        while(1) {
            $main_window_mutex->down();
            if ($main_window_inited == 1) {
                $main_window_mutex->up();
                last;
            }
            $main_window_mutex->up();
            # dormir por 0.1 segundos
            select(undef, undef, undef, 0.1);
        }
        
        # fechar outras threads
        fechar_tudo();
        print STDERR $erro;
        return;
    }

    # 'agora vamos comecar a putaria.'
   
    # conectar no servidor
    $asterisk_out->enqueue(freeze {"TYPE" => "LOG", "DATA" => "Connecting on manager..."});
    my $socket = IO::Socket::INET->new(PeerAddr => "$server:$port");

    if (not $socket) {
        $asterisk_out->enqueue(freeze {"TYPE" => "LOG", DATA => "Cannot connect on manager" });
        return;
    }
    $asterisk_out->enqueue(freeze {"TYPE" => "LOG", DATA => "Connected" });
    
    # ler boas vindas
    <$socket>;
    
    $asterisk_out->enqueue(freeze {"TYPE" => "LOG", DATA => "Sendding login information" });

    send_action($socket, {Action => "login", Username => $username, Secret => $secret});
    
    my $line;
    my %result;
    
    while ( defined($line = <$socket>) and $line !~ /^\s*$/) {
        # limpar '\r\n'
        $line =~ s/\s*$//g;

        my ($key, $value) = split(/:\s*/, $line, 2);
        $result{lc($key)} = lc($value);
    }
    if ($result{response} eq 'success') {
        $asterisk_out->enqueue(freeze {"TYPE" => "LOG", DATA => "Logged on manager" }); 
    } else {
        $asterisk_out->enqueue(freeze {"TYPE" => "LOG", DATA => "Fail to login on manager" });
        $socket->close();
        return;
    }
    undef($line);
    undef(%result);
   

    my $s = IO::Select->new($socket);
    my $internal_queue = Thread::Queue->new();
    my $buffer  = [];
    
    # entrar em loop de eventos
    while (not $quit) {
        if ( $s->can_read(0.1) ) {
            $socket->blocking(0);
            my @lines = <$socket>;
            foreach my $line (@lines) {
                if ($line =~ /^\s*$/) {
                    # fim de um comando
                    my %data;

                    foreach my $b (@{$buffer}) {
                        my ($key, $value) = split(/:\s*/, $b, 2);
                        $data{lc($key)} = $value;
                    }
                    
                    $internal_queue->enqueue(freeze(\%data));
                    $buffer = [];
                } else {
                    $line =~ s/\s+$//;
                    push (@{$buffer}, $line);
                }
            }

            $socket->blocking(1);
        }
        # eventos do usuario
        if (my $e = thaw($asterisk_in->dequeue_nb())) {
            #print Dumper($e);
            if ($e->{TYPE} eq 'QUEUE_ADD') {
                send_action($socket, {
                    Action => 'QueueAdd',
                    Queue => $queue,
                    Interface => "Agent/$agent" 
                });
                send_action($socket, {
                    Action => 'AgentCallBackLogin',
                    Agent => $agent,
                    Exten => $extension
                });
                #informar meus dados 
                $asterisk_out->enqueue(freeze {
                    "TYPE" => "INFO", 
                    DATA => {
                        agent => $agent,
                        extension => $extension,
                        queue => $queue,
                        channel => $channel
                    }
                });

            } elsif ($e->{TYPE} eq 'AGENT_PAUSE') {
                send_action($socket, {
                    Action => 'QueuePause',
                    Queue => $queue,
                    Interface => "Agent/$agent",
                    Paused => 1
                });
            } elsif ($e->{TYPE} eq 'AGENT_UNPAUSE') {
                send_action($socket, {
                    Action => 'QueuePause',
                    Queue => $queue,
                    Interface => "Agent/$agent",
                    Paused => 0
                });
            }
        }

        # eventos do manager
        if (my $e = thaw($internal_queue->dequeue_nb())) {
            #print Dumper($e); 
            if (
                exists($e->{event}) and
                lc($e->{event}) eq 'dial'
            ) {
                # ramal tocando
                my @temp = split("-", $e->{destination});
                # excluir do ultimo '-' em diante
                my $ramal = join("-", @temp[0..$#temp - 1]);
                
                # eh meu ramal?
                if ($ramal eq $channel) {
                    #print Dumper($e);
                    my $caller_id_name = $e->{calleridname};
                    my $caller_id = $e->{callerid};
                    
                    $asterisk_out->enqueue(
                        freeze {TYPE=>"BINA", DATA=>"\"$caller_id_name\" <$caller_id>"}
                    );
                }
            } elsif (
                exists($e->{event}) and
                lc($e->{event}) eq 'hangup'
            ) {
                # ramal desligado
                my @temp = split("-", $e->{channel});
                # excluir do ultimo '-' em diante
                my $ramal = join("-", @temp[0..$#temp - 1]);
                # eh o meu ramal?
                if ($ramal eq $channel) {
                    #print Dumper($e);
                    $asterisk_out->enqueue(freeze { TYPE => "CLEAN" });
                }

            } elsif (
                exists($e->{response})
            ) {
                $asterisk_out->enqueue(freeze { TYPE => "LOG", DATA => $e->{message}});
                if (index($e->{message}, "paused") >=0 ) {
                    $asterisk_out->enqueue(freeze { TYPE => "CLEAN" });
                }
            }
        }
    } 
    send_action($socket, { Action => "AgentLogoff", Agent => $agent});
    send_action($socket, { Action => "QueueRemove", Interface => "Agent/$agent", queue => $queue});
    send_action($socket, { Action => "logoff" });
    $socket->close();
    
}
sub pausar_despausar {

    #print Dumper(\@_);
    my $button = $_[1]->{"button_pausar"};
    #my $entry  = $_[1]->{"entry_bina"};

    if ($button->get_label() eq 'Login') {
        # queue add
        $asterisk_in->enqueue(freeze { TYPE => 'QUEUE_ADD' });
        $button->set_label('Pausar');
    } elsif ($button->get_label() eq 'Pausar') {
        # pausar
        $asterisk_in->enqueue(freeze { TYPE => 'AGENT_PAUSE' });
        $button->set_label('Despausar');
    } elsif ($button->get_label() eq 'Despausar') {
        # despausar
        $asterisk_in->enqueue(freeze { TYPE => 'AGENT_UNPAUSE' });
        $button->set_label('Pausar');
    }
}

sub configurar {

    #print Dumper(\@_);
    print "Not implemented\n";
}

sub fechar_tudo {
    #print Dumper(\@_);
    Gtk2->main_quit;
    $quit = 1;
    return 1;
}
sub send_action {
    my $socket = shift();
    my $h = shift() ;
    my @message;

    foreach my $key (keys(%{$h})) {
        push(@message, join(": ", $key, $h->{$key}));
    }
    print $socket join("\r\n", @message, "\r\n");
}

