#!/usr/bin/perl
# Implementar um proxy que nao eh proxy
# FIXME: usar ActionID
#

use strict;
use warnings;
use threads;
use threads::shared;

use File::Basename;
use Config::Tiny;
use IO::Socket::INET;
use IO::Select;
use Thread::Queue;
use Data::Dumper; 
use Storable qw(freeze thaw);
use Tie::IxHash;

#my $debug : shared = 1;
my $debug = 1;

# comandos permitidos pelos clientes
# 1 => passar para o asterisk sem tratamento
# 0 => tratar localmente
my %commands = ( 
    agentcallbacklogin => 1,
    queueadd => 1,
    queueremove => 1,
    queuepause => 1,
    queuestatus => 1,
    status => 1,
    ping => 0,
    login => 0,
    logoff => 0,
);
    
# flag para sair das threads;
our $quit :shared = 0;

# capturar o CTRL + C
#sub ctrl_c { 
#    print "ok\n"; 
#    $quit = 1;
#    $SIG{INT} = \&ctrl_c;
#}
#$SIG{INT} = \&ctrl_c;

# arquivo de confifuracao
my $proxy_conf_path = join("/", dirname($0) , "../cfg/proxy.ini");

# thread do manager
my $thr_manager;

# thread para despachar os eventos para os clientes
my $thr_events;

# thread para registrar conexoes
my $thr_connections;

# filas para comunicacao
my $thr_manager_in_queue = Thread::Queue->new();
my $thr_manager_out_queue = Thread::Queue->new();

my $thr_connections_in_queue = Thread::Queue->new();
my $thr_connections_out_queue = Thread::Queue->new();

$thr_manager = threads->new(\&thr_manager_func);
$thr_events = threads->new(\&thr_events_func);
$thr_connections = threads->new(\&thr_connections_func);

# esperar minhas threads

$thr_manager->join();
$thr_events->join();
$thr_connections->join();


###############################################
# thread do manager
# funcao: le eventos do manager, postar acoes no manager, nada mais
# 
sub thr_manager_func {
    debug_print('M', "Init");

    # ler arquivo de configuracao
    my $Config = Config::Tiny->read( $proxy_conf_path );
    
    my $server = $Config->{MANAGER}->{server} || "localhost";
    my $port = $Config->{MANAGER}->{port} || "5038";
    my $username = $Config->{MANAGER}->{username} 
        or die "no username in '$proxy_conf_path' [MANAGER]";
    my $secret = $Config->{MANAGER}->{secret} 
        or die "no secret in '$proxy_conf_path' [MANAGER]";
    
    debug_print(
        'M', 
        "Connecting to host '$server' on port '$port'", 
        "with username=$username"
    );
        
    my $socket = IO::Socket::INET->new(PeerAddr => "$server:$port")
        or die "Cannot connect to host '$server' on port '$port'";
    
    # ler mensagem de boas vindas
    <$socket>;
    
    # fazer o login
    my $login = {
        "Action" => "login",
        "Username" => $username,
        "Secret" => $secret
    };

    debug_print('M', "Logging...");
    
    send_action($socket, $login);
 
    my %result; 
    my $line;
    while ( defined($line = <$socket>) and $line !~ /^\s*$/) {
        # limpar '\r\n'
        $line =~ s/\s*$//g;

        my ($key, $value) = split(/:\s*/, $line, 2);
        $result{lc($key)} = lc($value);
    }
    if ($result{response} eq 'success') {
        debug_print('M', "Logged");
    } else {
        die "Cannot log in asterisk manager";
    }

    undef(%result);
    undef($line);

    # esperar por eventos vindos do asterisk e despachar para thr_events
    # como tambem actions vindos thr_events e despachar para o asterisk
    my $s = IO::Select->new($socket);
    my $buffer = [];

    while (not $quit) {
        if ( $s->can_read(0.1) ) {
            $socket->blocking(0);
            my @lines = <$socket>;
            foreach my $line (@lines) {
                if ($line =~ /^\s*$/) {
                    # fim de um comando
                    tie my %data, 'Tie::IxHash';

                    foreach my $b (@{$buffer}) {
                        my ($key, $value) = split(/:\s*/, $b, 2);
                        $data{lc($key)} = $value;
                    }
                    #debug_print("M", "New event from asterisk", Dumper(\%data));
                    $thr_manager_out_queue->enqueue(freeze(\%data));
                    $buffer = [];
                } else {
                    $line =~ s/\s+$//;
                    push (@{$buffer}, $line);
                }
            }

            $socket->blocking(1);
        }
        # eventos from thr_events
        while (my $event = thaw($thr_manager_in_queue->dequeue_nb())) {
            debug_print("M", "New event from thr_events");
            send_action($socket, $event);        
        }
    } 
    debug_print('M', "End");
}

###############################################################################
# thread para tratar eventos
sub thr_events_func {
    debug_print('E',"Init");
    my @fifo_responses; 
    my $waiting_for_more = 0;
    my %logged;

    my $Config = Config::Tiny->read($proxy_conf_path);
    # IMPORTANT: ficar so com os usuarios
    # se adicionar mais uma secao especial no XXX.ini, remover aqui
    delete ($Config->{SERVER});
    delete ($Config->{MANAGER});
    
    while (not $quit) {
        while(my $data = thaw($thr_manager_out_queue->dequeue_nb())) {
            debug_print('E', "new data from manager", Dumper($data));
            if (exists($data->{event})) {
                if ($waiting_for_more) {
                    $data->{ID} = $fifo_responses[0];
                    $thr_connections_in_queue->enqueue(freeze($data));
                    if ($data->{event} =~ /.Complete$/) {
                        $waiting_for_more = 0;
                        shift(@fifo_responses);
                    }
                # avisar de bina
                } elsif (
                    #lc($data->{event}) eq 'newchannel' and 
                    #lc($data->{state}) eq 'ringing'
                    lc($data->{event}) eq 'dial' or
                    lc($data->{event}) eq 'hangup'
                ) {
                    foreach my $ID (keys(%logged)) {
                        $data->{ID} = $ID;
                        $thr_connections_in_queue->enqueue(freeze($data));
                    }
                }
            } elsif (exists($data->{response})) {
                $data->{ID} = shift(@fifo_responses);
                $thr_connections_in_queue->enqueue(freeze($data));
                if (exists($data->{message}) and $data->{message} =~ /.will follow$/) {
                    # ainda nao acabou
                    $waiting_for_more = 1;
                    unshift(@fifo_responses, $data->{ID});
                }
            }
        }
        while(my $data = thaw($thr_connections_out_queue->dequeue_nb)) {
            debug_print('E', "new data from client " . $data->{ID} , Dumper($data));
            # FIXME: capturar logins (eu faco os logins, nao o asterisk)
            if (not exists($data->{"action"})) {
                tie my %er_data, 'Tie::IxHash';
                $er_data{ID} = $data->{ID};
                $er_data{response} = 'Error';
                $er_data{message} = 'Missing action in request';
                $thr_connections_in_queue->enqueue(freeze(\%er_data));
            # comandos permitidos via passthru...
            } elsif (
                exists $logged{$data->{ID}} and
                exists $commands{lc($data->{"action"})} and 
                $commands{lc($data->{"action"})} == 1
            ) {
                # enviar para o manager, guardar ordem
                push(@fifo_responses, $data->{ID});
                # nao passar ID
                delete($data->{ID});
                $thr_manager_in_queue->enqueue(freeze($data));
            } elsif (exists $commands{lc($data->{"action"})} and $commands{lc($data->{"action"})} == 0) {
                tie my %r_data, 'Tie::IxHash';
                $r_data{ID} = $data->{ID};
                my $action = $data->{"action"};
                if (lc($action) eq 'login') {
                    # fazer login
                    if (not exists($data->{username}) or not exists($data->{secret})) {
                        $r_data{response} = "Error";
                        $r_data{message} = "Authentication failed";
                    } else {
                        if (
                            exists $Config->{$data->{username}} and 
                            $Config->{$data->{username}}->{secret} eq $data->{secret}
                        ) {
                            $r_data{response} = "Success";
                            $r_data{message} = 'Authentication accepted';
                            $logged{$data->{ID}} = 1;
                        } else {
                            $r_data{response} = "Error";
                            $r_data{message} = "Authentication failed";
                        }
                    }
                } elsif (lc($action) eq 'ping' ) {
                    $r_data{response} = 'Pong';
                } elsif (lc($action) eq 'logoff') {
                    $r_data{response} = 'Success';
                    $r_data{message} = 'See you later';
                    delete $logged{$data->{ID}} if exists $logged{$data->{ID}};
                } 
                $thr_connections_in_queue->enqueue(freeze(\%r_data));
            } else {
                tie my %er_data, 'Tie::IxHash';
                $er_data{ID} = $data->{ID};
                $er_data{response} = 'Error';
                $er_data{message} = 'Invalid/unknown command | Permission denied';
                $thr_connections_in_queue->enqueue(freeze(\%er_data));
            }
        }
       
        # dormir(sleep) por 0.1 segundos
        select(undef, undef, undef, 0.1);
    }

    debug_print('E',"End");
}

###############################################################################
# thread para registrar conexoes
sub thr_connections_func {
    debug_print('C',"Init");
    
    # ler arquivo de configuracao
    my $Config = Config::Tiny->read($proxy_conf_path);
    
    my $srv_socket = IO::Socket::INET->new(
        Listen => 1,
        LocalAddr => $Config->{SERVER}->{bind} || "0.0.0.0",
        LocalPort => $Config->{SERVER}->{port} || 9987,
        Proto => 'tcp',
        ReuseAddr => 1,
    ) or die "Cannot create server: $@";
    
    debug_print('C', "Listening for connections on " . 
        join(":", $srv_socket->sockhost(), $srv_socket->sockport)
    );
    my %clients;

    my $sel = new IO::Select($srv_socket);
    while (not $quit) {
        while (my @ready = $sel->can_read(0.1)) {
            foreach my $socket (@ready) {
                if ($socket == $srv_socket) {
                    # nova conexao
                    my $client_socket = $srv_socket->accept();
                    my $id = join(":", $client_socket->peerhost(), $client_socket->peerport());
                    $clients{$id} = { 
                        "socket"    => $client_socket,
                        "buffer"    => [],
                    };
                    debug_print('C', "new connection from: " .  $id);
                    print $client_socket "Obelisk (Asterisk Proxy)/0.1\r\n";
                    $sel->add($client_socket);
                } else {
                    my $id = join(":", $socket->peerhost(), $socket->peerport());

                    if ($socket->connected() and not $socket->eof() ) {
                        $socket->blocking(0);
                        
                        my @lines = <$socket>;
                        
                        foreach my $line (@lines) {
                            if ($line =~ /^\s*$/) {
                                # fim de um comando
                                #debug_print("C", "client $client_id sends", @{$client->{buffer}});
                                #my %data;

                                tie my %data, 'Tie::IxHash';
                                
                                foreach my $b (@{$clients{$id}->{buffer}}) {
                                    my ($key, $value) = split(/:\s*/, $b, 2);
                                    $data{lc($key)} = $value;
                                }
                                $data{ID} = $id;
                                $thr_connections_out_queue->enqueue(freeze(\%data));
                                
                                $clients{$id}->{buffer} = [];
                            } else {
                                $line =~ s/\s+$//;
                                push (@{$clients{$id}->{buffer}}, $line);
                            }
                        }
                        $socket->blocking(1);
                    } else {
                        debug_print('C', "Client $id disconnected");
                        
                        my %data;
                        $data{ID} = $id;
                        $data{action} = "logoff";
                        $thr_connections_out_queue->enqueue(freeze(\%data));
                        
                        $sel->remove($socket);
                        delete($clients{$id});
                        $socket->close();
                    }
                }
            }
        }
        while(my $data = thaw($thr_connections_in_queue->dequeue_nb)) {
            if (exists($clients{$data->{ID}})) {
                my $socket = $clients{$data->{ID}}->{socket};
                # apagar ID
                delete($data->{ID});
                send_action($socket, $data);
            }
        }
    }
    debug_print('C',"End");
}



# funcoes auxiliar

sub debug_print {
    my $category = shift();
    my $first_message = shift();
    chomp($first_message);
    my @messages = @_;

    my $output;

    $first_message = "> " . $first_message if (@messages);
    
    $output .= "[$category] " . $first_message . "\n";
    
    foreach my $m (@messages) {
        chomp($m);
        $output .= " " x length("[$category]") . " > " . $m . "\n";
    }
    print STDERR $output if $debug;
}

# envia dados para o manager
sub send_action {
    my $socket = shift();
    my $h = shift() ;
    my @message;

    foreach my $key (keys(%{$h})) {
        push(@message, join(": ", $key, $h->{$key}));
    }
    print $socket join("\r\n", @message, "\r\n");
}


